// control.c
// control routines for open-source proportional solenoid valve ventilator
// (c) Bill Dally, April 2020
//
// Copyright 2020 Bill Dally
// 
// Licensed under the Ventilator Open Source License (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// ./LICENSE
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include "main.h"
#include "solenoid.h"
#include "alarm.h"
#include "ui.h"

uint8_t first ;

// proportional and derivative errors
void compute_flow_errors(struct pid *pid_p, int16_t flow, int16_t mflow) {
  pid_p->o2err = pid_p->oerr ;
  pid_p->oerr = pid_p->err ; // for diff
  pid_p->err = flow - mflow ; // error signal
  // smooth d_err - two step
  pid_p->derr = (pid_p->err - pid_p->oerr) + (pid_p->err - pid_p->o2err) ;
}

//**
void compute_press_errors(int16_t setp) {
  opress_err = press_err ; // for derivative
  press_err = setp - mpressure ; //
  press_derr = press_err - opress_err ;
}

// clamp integral error
void limit_i_err(struct pid *pid_p) {
  if(pid_p->ierr > MAX_IERR) pid_p->ierr = MAX_IERR ;
  if(pid_p->ierr < MIN_IERR) pid_p->ierr = MIN_IERR ;
}

// version for pressure ierr
void limit_i_err_p(struct pid *pid_p) {
  if(pid_p->ierr > MAX_PIERR) pid_p->ierr = MAX_PIERR ;
  if(pid_p->ierr < MIN_PIERR) pid_p->ierr = MIN_PIERR ;
}

// **
void limit_cflow(struct pid *pid_p) {
  if(pid_p->cflow < 0) pid_p->cflow = 0 ;
  if(pid_p->cflow > MAX_FLOW_MLPS) pid_p->cflow = MAX_FLOW_MLPS ;
}


// PID control law with feed forward
// call with ptr to pid struct, target, and measured flow
// commanded flow is returned in pid struct
void pid_ff(struct pid *pid_p, uint16_t flow, uint16_t mflow) {
  compute_flow_errors(pid_p, flow, mflow) ;
  pid_p->ierr = pid_p->ierr + pid_p->err ;
  limit_i_err(pid_p) ;

  // derivative error is subtracted
  lcflow = (int32_t) flow + (((int32_t)pid_p->err * pgain 
			      +  (int32_t) pid_p->ierr * igain 
			      - (int32_t) pid_p->derr * dgain) >> GAIN_DIV_LG2) ;
  pid_p->cflow = lcflow ;
  limit_cflow(pid_p) ;
}

// same as above but with pressure gains
// old version - same gain both ways
void pid_ff_ip(struct pid *pid_p, uint16_t flow, uint16_t mflow) {
  compute_flow_errors(pid_p, flow, mflow) ;
  pid_p->ierr = pid_p->ierr + pid_p->err ;
  limit_i_err(pid_p) ;
  lcflow = (int32_t) flow + (((int32_t)pid_p->err * press_gain +
			      (int32_t) pid_p->ierr * press_igain +
			      (int32_t) pid_p->derr * press_dgain) >> GAIN_DIV_LG2) ;
  pid_p->cflow = lcflow ;
  limit_cflow(pid_p) ;
}

// same as above but with pressure gains
// new version - integral more aggressive if error is negative - and above fuzz
void pid_ff_p(struct pid *pid_p, uint16_t flow, uint16_t mflow) {
  compute_flow_errors(pid_p, flow, mflow) ;
  if(pid_p->err < PDOWN_MARGIN) {
    // extra effort when above limit pressure by margin
    pid_p->ierr = pid_p->ierr + PDOWN_GAIN*pid_p->err ; 
  } else if (pid_p->err < PDOWN_MARG2) {
    // even more effort when higher above limit pressure
    pid_p->ierr = pid_p->ierr + PDOWN_GAIN2*pid_p->err ; 
  } else if (pid_p->err < PDOWN_MARG3) {
    // and when way over really ramp down
    pid_p->ierr = pid_p->ierr + PDOWN_GAIN3*pid_p->err ; 
  } else {
    pid_p->ierr = pid_p->ierr + pid_p->err ;
  }
  limit_i_err_p(pid_p) ;

  // note that derr is subtracted
  lcflow = (int32_t) flow + (((int32_t)pid_p->err * press_gain 
			      + (int32_t) pid_p->ierr * press_igain 
			      - (int32_t) pid_p->derr * press_dgain) >> GAIN_DIV_LG2) ;
  pid_p->cflow = lcflow ;
  limit_cflow(pid_p) ;
}


/*
void compute_press_flow(uint8_t cap_flow) {
  i_err = i_err + (press_err * press_igain) ; // integrate pressure error - with press_igain.  Use regular igain below.
  limit_i_err() ;
  if(cap_flow == 1) {
    lcflow = ((int32_t)press_err * press_gain + (int32_t) ((i_err * igain)>>1) +
	      (int32_t) press_derr * press_dgain) >> GAIN_DIV_LG2 ;
  } else {
    lcflow = ((int32_t)press_err * press_gain + (int32_t) i_err * igain +
	      (int32_t) press_derr * press_dgain) >> GAIN_DIV_LG2 ;
  }
  cflow = lcflow ;
  limit_cflow() ;
  }

// PID control for pressure only
int16_t pid_press(uint16_t setp, uint16_t mpressure, uint8_t cap_flow) {
  compute_press_errors(setp) ;
  compute_press_flow(cap_flow) ;
  if(cap_flow && (cflow > CAP_FLOW)) cflow = CAP_FLOW ;
  return(cflow) ;
}*/

// one-sided pressure for v3.0 vent
// returns pid_p->cflow
// if pressure below guard band return target flow
// 
void pid_ff_press(struct pid *pid_p, uint16_t flow, int16_t maxp, int16_t mpressure) {
  if(mpressure <= (maxp - PGUARD)) { // return target flow for flow regulation
    pid_p->cflow = flow ;
   } else if((mpressure > (maxp - PGUARD)) && (mpressure < maxp)) { // guard band -
    // blend cflow based on fraction
    pid_ff_p(&press_pid, maxp, mpressure) ;
    // compute fraction 
    // frac is fraction of normal flow, fracx is fraction of pressure computed flow
    // use next line to zero last half of guard band
    //frac = (maxp - mpressure) - HALF_PGUARD ; // numerator - denom is PGUARD - 
    frac = (maxp - mpressure) ; // numerator - denom is PGUARD
    if(frac < 0) frac = 0 ;
    fracx = PGUARD - frac ;
    // interpolate
    int32_t lcflow = (((int32_t) pid_p->cflow * fracx) + ((int32_t) flow * frac)) >> PGUARD_LG2 ;  
    pid_p->cflow = lcflow ;
  } else { // regulate pressure
    pid_ff_p(&press_pid, maxp, mpressure) ;
  }
}

//PID control law with feed forward and one-sided pressure
/*
int16_t pid_ff_press(uint16_t flow, uint16_t mflow, uint16_t maxp, uint16_t mpressure) {
  compute_flow_errors() ;
  compute_press_errors(maxp) ;
  lcflow = (int32_t) flow + (((int32_t)error*pgain + (int32_t) d_err*dgain) >> GAIN_DIV_LG2 ) ; // less igain
  if(mpressure <= (maxp - PGUARD)) { // regulate flow - ** add derivative of pressure to look ahead
    cmode = 0 ;
    i_err = i_err + error ;// identical to above
    limit_i_err() ;
    lcflow += ((int32_t) i_err * igain) >> GAIN_DIV_LG2 ; 
  } else if((mpressure > (maxp - PGUARD)) && (mpressure < maxp)) { // guard band - ramp down flow integral error
    if(mflow < (flow>>1)) { // at low flows, just regulate pressure
      cmode = 4 ;
      compute_press_flow(0) ;
    } else {
      cmode = 1 ;
      frac = (maxp - mpressure) - HALF_PGUARD ; // numerator - denom is PGUARD - zero in last half of guard band
      if(frac < 0) frac = 0 ;
      int32_t lferr = ((int32_t) error * frac) >> PGUARD_LG2 ;  // scale by frac/PGUARD
      ferr = lferr ;
      i_err = i_err + ferr ;
      limit_i_err() ;
      lcflow += ((int32_t) i_err * igain) >> GAIN_DIV_LG2 ;
    }
  }else { // regulate pressure
    cmode = 2 ;
    compute_press_flow(0) ;
  }
  cflow = lcflow ;
  limit_cflow() ;
  //put_word(flow); comma() ; put_word(mflow) ; comma() ;
  //put_signed(error) ; comma() ; put_signed(cflow); crlf() ;
  return(cflow) ;
}
*/

/*
// vestigial routine - no longer used
void volume_control(void) {
  if(time == 0) { // first cycle, just feed forward
    tflow = flow_step ;
    i_err = 0 ; // zero integral error
    mvol = 0 ; // zero measured volume each breath
    cflow = pid_ff(tflow, mflow) ;
  } else if(time < trise) {
    i_err = 0 ; // avoid windup during transient
    tflow = tflow + flow_step ;
    cflow = pid_ff(tflow, mflow) ;
    tdown = time ; // for next segment
  } else if(mvol <= (vol_limit - ramp_vol)) {
    tflow = flow ;
    cflow = pid_ff(tflow, mflow) ;
    tdown = time ; // for next segment
  } else { // within ramp of target volume
    i_err = 0 ; // or perhaps reduce gain
    uint16_t tleft = trise - (time - tdown) ; // time left in ramp
    uint32_t rramp_vol = (tflow * tleft)/ (2 * TICK) ; // remaining ramp volume
    if(mvol >= (vol_limit - rramp_vol)) tflow = tflow - flow_step ;
    if (mvol >= vol_limit) tflow = 0 ; // cut to zero if at target volume
    cflow = pid_ff(tflow, mflow) ;
    if(tflow <= 0 ) {
      tflow = 0 ;
      cflow = 0 ;
    } else {
      cflow = pid_ff(tflow, mflow) ;
    }
  }
  if(time > (ton + trise*4)) cflow = 0 ;
}
*/

void expiratory_control() {
  if(first == 1) {
    toff = time ;
    first = 0 ;
    vestate = VE_START ;
    spc_high = 0 ;
    ps_timer = 0 ; // used here to sample platp
  }
  if(vestate == VE_START) { // start of expiratory period
    ps_timer += TICK ;
    if(ps_timer < INSPIRATORY_SAMPLE_TIME) platp = smpressure ;
    cflow = 0 ;
    ca_flow = 0 ;
    co_flow = 0 ;
    sol3_off() ;
    i_err = 0 ;
    // start regulating peep
    if((smpressure <= (set_peep_close)) && (set_peep > 0) && (disconnect_count == 0)) {
      vestate = VE_REG ;
      sol3_on() ;
    }
  } else if(vestate == VE_REG) {
    // this code regulated pressure - idea is not to regulate, just close exhale valve at right time
    //if(sol3_state == 1) cflow = pid_press(set_peep, mpressure, 1) ; // pressure control on PEEP if valve on
    //else i_err = 0 ; // start from 0 - don't integrate with valve off
    if(smpressure > set_peep) {
      i_err = 0 ; //**
      cflow = 0 ;
      ca_flow = 0 ;
      co_flow = 0 ;
    }
    // only open if spc is pegged
    if((smpressure > set_peep_open) && (set_peep_close <= SPC_INC)) {
      sol3_off() ;
      spc_high = 1 ;
    }
    // don't turn back on if PEEP is 0 or disconnect
    if((smpressure < set_peep_close) && (set_peep > 0) && (disconnect_count > 0)) { 
      sol3_on() ;
    }
  }
}

// splits cflow into ca_flow and co_flow
void split_flow(int16_t cflow) {
  int32_t la_flow = ((int32_t) cflow * (int32_t)frac_air_x1k) >> 10 ;
  ca_flow = la_flow ;
  int32_t lo_flow = ((int32_t) cflow * (int32_t)frac_o2_x1k) >> 10 ;
  co_flow = lo_flow ;
}
 
// simple time based control
void control(void) {
  if(time == 0) {
    mvol = 0 ; // zero measured volume at start of each breath
    first = 1 ;  // to capture first cycle at end of vol
    tflow = flow_step ;
    if(mode != MODE_BLEND) sol3_on() ; // exhale valve closes
  }
  if(mode == MODE_BLEND) {
    cflow = bflow ;
    split_flow(cflow) ;
    pid_ff(&air_pid, ca_flow, ma_flow) ; // ca here is target
    ca_flow = air_pid.cflow ;
    pid_ff(&o2_pid, co_flow, mo_flow) ; // co here is target
    co_flow = o2_pid.cflow ;
  } else if((mode == MODE_VOL)||(mode == MODE_VOL_ASSIST)) { // volume control mode
    if(time < TIME_FF) { // first cycles, just feed forward
      sol3_on() ;
      if(time < trise) { // ramp for rise time, then steady
	tflow = tflow + flow_step ;
      } else  {
	tflow = flow ;
      }
      if(smpressure <= maxp) {
	split_flow(tflow) ;
      } else {
	ca_flow = 0 ;
	co_flow = 0 ;
      }
      vestate = VE_START ;
      air_pid.ierr = 0 ;  // zero three integral errors
      o2_pid.ierr = 0 ;
      press_pid.ierr = 0 ; 
      pid_ff(&air_pid, a_flow, ma_flow) ; // prime d_err
      pid_ff(&o2_pid, o_flow, mo_flow) ; // and for o2
    } else if((time < ton) || ((mvol < vol_limit) && (time < tmax))) { // PID control for remainder of breath
      // pressure control computes target flow - one sided
      sol3_on() ;
      pid_ff_press(&press_pid, flow, maxp, mpressure) ; // use pressure to update target flow
      split_flow(press_pid.cflow) ;
      // at this point ca an co_flow are target flows - possibly reduced for pressure
      pid_ff(&air_pid, ca_flow, ma_flow) ;
      ca_flow = air_pid.cflow ;
      pid_ff(&o2_pid, co_flow, mo_flow) ;
      co_flow = o2_pid.cflow ;
    } else if(time < (ton + thold - INSPIRATORY_PAUSE_TIME)) { // inspiratory hold - no flow sol3 on
      sol3_on() ;
      co_flow = 0 ;
      ca_flow = 0 ;
      cflow = 0 ;
    } else { // expiratory period
      expiratory_control()  ;
    }

  } else { // pressure control mode -
    // PRESSURE
    if(time == 0) {
      press_pid.ierr = PS_INIT_IERR ; // init pressure integral error
      //pstate = PS_RISING ;
      pstate = PS_RISE4 ; // Closed loop control works great from the start
      ps_timer = PS_DELAY ;
      sol3_on() ;
    }
    if(mvol > vol_limit) { // cut off when volume limit is reached
      pstate = PS_IDLE ;
    }
    // if pressure control mode, terminate on time
    if((mode == MODE_PRESS) && (time > ton)) {
      pstate = PS_IDLE ;
    }
    // if no pressure rise in alarm time - shut down
    if((pstate != PS_IDLE) && (time > PRESS_ALARM_TIME) && (smpressure < (peep + pip_disconnect))) {
      alarm_trig(ALARM_DISCONNECT) ;
      stop_vent() ;
    }
    if(pstate == PS_RISING) { // PS Flow until within guardband of target pressure
      cflow = PS_FLOW ;
      if(smpressure > (maxp - PS_GUARDBAND)) { // use smoothed pressure
	pstate = PS_RISE2 ;
        press_pid.ierr = PS_INIT_IERR ; // start fresh
      }
    } else if (pstate == PS_RISE2) { // then ramp flow to PS_FLOW2
      cflow = cflow - PS_FLOW_DEC ;
      if((cflow < PS_FLOW2) ||
	 //(smpressure < osmpressure) ||
	 (smpressure > (maxp - PS_GB2))) {
	pstate = PS_RISE3 ;
      }
    } else if (pstate == PS_RISE3) { // Then regulate until target pressure achieved and time elapsed
      cflow = PS_FLOW2 ; // if not yet at target pressure or guardband run until there
      if((smpressure < osmpressure) || (smpressure > (maxp - PS_GB2))) {
	pstate = PS_RISE4 ;
	ps_timer = PS_DELAY ;
      }
      
    } else if (pstate == PS_RISE4) { // Then regulate until target pressure achieved and time elapsed
      // This is where things really start
      if(smpressure < (maxp - PS_GUARDBAND)) { // zero flow integral error until time
	air_pid.ierr = 0 ;
	o2_pid.ierr = 0 ;
      }
      pid_ff_p(&press_pid, maxp, mpressure) ;
      cflow = press_pid.cflow ;
      //cflow = pid_press(maxp, mpressure, 0) ; // straight pressure control
      if(ps_timer > 0) ps_timer -= TICK ;
      if((smpressure >= (maxp - PS_FUZZ)) && (ps_timer == 0)) {
	pstate = PS_HOLD ;
	ps_timer = PS_DELAY2 ;
      }
    } else if (pstate == PS_HOLD) {
      pid_ff_p(&press_pid, maxp, mpressure) ;
      cflow = press_pid.cflow ;
      //cflow = pid_press(maxp, mpressure, 0) ; // straight pressure control
      if(ps_timer > 0) ps_timer -= TICK ;
      if(ps_timer == 1) {
	ps_flow = mflow ; // record peak flow
	ps_off_flow = (ps_flow * ps_frac) >> PS_LG_DENOM ; // fraction is PS_FRAC/PS_SHIFT - nom 50%
      }
      // in assist mode turn off on flow
      if((ps_timer == 0) && (mode == MODE_PRESS_ASSIST)) { // in hold period proper
	//if(mflow > ps_flow) { // reset if too low
	//  ps_flow = mflow ;
	//  ps_off_flow = (ps_flow * ps_frac) >> PS_LG_DENOM ; // fraction is PS_FRAC/PS_SHIFT - nom 50%
	//}
	if(mflow <= ps_off_flow) {
	  pstate = PS_IDLE ;
	  vestate = VE_START ;
	  cflow = 0 ;
	}
      }
    } else { // PS_IDLE
      cflow = 0 ;
      expiratory_control() ;
      /* old control
      cflow = 0 ;
      if(smpressure > set_peep) {
	sol3_off() ;
      }	else {
	sol3_on() ;
      }
      */
    }
    // in all pressure modes, cut off flow if above
    /*
    if(smpressure > (maxp + PDOWN_CUTOFF)) {
      cflow = 0 ;
      }
    */
    // in all pressure modes split cflow and apply control to air and o2
    if(cflow == 0) {
      ca_flow = 0 ;
      co_flow = 0 ;
    } else {
      split_flow(cflow) ;
      pid_ff(&air_pid, ca_flow, ma_flow) ; // ca here is target
      ca_flow = air_pid.cflow ;
      pid_ff(&o2_pid, co_flow, mo_flow) ; // co here is target
      co_flow = o2_pid.cflow ;
    }
  }
} // ca_flow and co_flow out of here are applied in callback in main.c
