// leds.c
// Routines to drive front-panel LEDs
//
// (c) Bill Dally, April 2020
//
// Copyright 2020 Bill Dally
// 
// Licensed under the Ventilator Open Source License (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// ./LICENSE
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

#include <avr/io.h>
#include "main.h"

volatile uint16_t led_timer = 0 ;

void set_leds(uint8_t val) { // bottom 3 bits to LEDs
  PORTC = (PORTC & 0x1F) | ((val & 0x7) << 5) ;
}

// LED 3 - Power
// Blink 100ms every 1s when on
// Steady on during inspiratory period
// Blink twice per second during expiratory period
// LED 2 - Mode
// Blink 100ms every 1s - out of phase with LED 1 for volume mode
// Blink twice for volume mode assist
// Steady on for pressure mode
// 50% duty factor for pressure mode assist
// Steady on for constant flow mode - see power 
// LED 1 - Alarm
// PC 7, 6, 5

void update_leds(void) {
  uint8_t val = 1 ;
  led_timer += 1 ; // bump timer every 1ms (called every 16ms)
  if((led_timer & 0x7F) == 1) { // once every 128ms
    uint8_t flash_timer = (led_timer >> 7) & 0x7 ; // count to 8
    // LED 3
    if(flash_timer == 0) val |= 4 ;
    if(run == 1) {
      if(time < ton) val |= 4 ; // inspiratory period
      else if(flash_timer == 2) val |= 4 ;
    }
    // LED 2
    switch(mode) {
    case MODE_VOL:
      if(flash_timer == 4) val |= 2 ;
      break ;
    case MODE_PRESS:
      val |= 2 ;
      break ;
    case MODE_VOL_ASSIST:
      if((flash_timer == 3) | (flash_timer == 5)) val |= 2 ;
      break ;
    case MODE_PRESS_ASSIST:
      if(flash_timer > 3) val |= 2 ;
      break ;
    case MODE_BLEND: // mode light steady on - power flashes double rate - extra flash when on
      val |= 2 ;
      if((flash_timer == 4)||(flash_timer == 0)) val |= 4 ;
      else val &= ~4 ;
      if((run == 1) && (flash_timer == 2)) val |= 4 ;
      break ;
    }
    // LED 1
    if(alarm == 0) val &= ~1 ;
    set_leds(val) ;
  }
}
