// ui_test.c
// test portion of user interface for open-source proportional solenoid valve ventilator
//
// (c) Bill Dally, April 2020
//
// Copyright 2020 Bill Dally
// 
// Licensed under the Ventilator Open Source License (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// ./LICENSE
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Enter Test UI by typing TES.  Front panel is locked out
//
#include <stdio.h>
#include <avr/interrupt.h>
#include "main.h"
#include "params.h"
#include "serial.h"
#include "flow.h"
#include "siargo.h"
#include "pwm.h"
#include "nxp_pressure.h"
#include "honeywell_pressure.h"
#include "honeywell_flow.h"
#include "mux.h"
#include "adc.h"
#include "lcd.h"
#include "control.h"
#include "alarm.h"
#include "ui.h"
#include "solenoid.h"
#include "watchdog.h"
#include "leds.h"
#include "restore.h"
#include "two_wire.h"

void test_v3() {
  uint16_t val ;
  uint8_t b ;
  int8_t i, done = 0 ;
  do {
    put_string_P(PSTR("Test V3:\n\r")) ;
    char c = serial_get() ;
    switch(c) {
    case 'd':
      watchdog_disable() ;
      cli() ;
      break ;
    case 'e':
      watchdog_init() ;
      sei() ;
      break ;
    case 'f':
      val = honeywell_get_flow() ;
      put_string_P(PSTR("flow = ")) ; put_word(val) ; crlf() ;
      break ;
    case 'h':
      for(i = 0 ; i < 100 ; i++) {
	val = honeywell_get_pressure2() ;
	put_word(val) ; space() ;
      }
      break ;
    case 'm':
      val = get_val() ;
      b = val ;
      set_flow_mux(b) ;
      put_string_P(PSTR("mux = ")) ; put_byte(b) ; crlf() ;
      break ;
    case 'p':
      val = get_val() ;
      put_string_P(PSTR("PWM1 = ")); put_word(val) ; crlf() ;
      set_pwm0(val) ;
      break ;
    case 'q':
      val = get_val() ;
      put_string_P(PSTR("PWM2 = ")); put_word(val) ; crlf() ;
      set_pwm1(val) ;
      break ;
    case 'Q':
      done = 1 ;
      break ;
    case '?':
    default:
      put_string_P(PSTR("d - disable interrupts\n\r")) ;
      put_string_P(PSTR("e - enable interrupts\n\r")) ;
      put_string_P(PSTR("f - get flow\n\r")) ;
      put_string_P(PSTR("h - get hp2 pressure\n\r")) ;
      put_string_P(PSTR("m - set mux\n\r")) ;
      put_string_P(PSTR("p - set pwm\n\r")) ;
      put_string_P(PSTR("q - set pwm2\n\r")) ;
      put_string_P(PSTR("Q - quit\n\r")) ;
    }
  } while(done == 0) ;
}

void test_ui() {
  uint16_t val, val2, i ; // temp
  uint32_t addr ;
  int8_t done = 0 ;
  do {
    put_string_P(PSTR("Test Interface:\n\r")) ;
    char c = serial_get() ;
    switch(c) {
    case 'a': // test wrap of gtime
      val = get_val() ; // amount past wrap
      val2 = get_val() ; // amount before wrap
      i = less_than_minute(val, val2) ;
      put_word(val) ; space() ; put_word(val2) ; space() ; put_word(i) ; crlf() ;
      break ;
    case 'b': // turn sol3 on
      sol3_on() ;
      break ;
    case 'c': // turn sol3 off
      sol3_off() ;
      break ;
    case 'd': // save parameters
      save_parameters(EER_OFFSET_1) ; // save at 32
      break ;
    case 'e': // read 16 words from EE Prom
      val = get_val() ;
      for(i=0; i<16;i += 2) {
	addr = val ;
	val2 = read_at_word(addr) ;
	put_word(val); space() ; put_word(val2) ; space() ;
	put_byte(val2 & 0xFF) ; space() ; put_byte(val2 >> 8) ; crlf() ;
	val += 2 ;
      }
      break ;
    case 'f':
      val = get_val() ;
      val2 = flow_to_pwm(val, 0) ; 
      put_word(val) ; dot() ; put_word(val2) ; crlf() ;
      break ;
    case 'g': // restore parameters
      restore_parameters(EER_OFFSET_1) ;
      break ;
    case 'h': // write EE Prom
      val = get_val() ; // address
      addr = val ;
      val2 = get_val() ; // data
      write_at_word(addr, val2) ;
      break ;
    case 'k': // zero pressure flow log
      pf_log_cnt = 0 ;
      pf_log_ptr = 0 ;
      break ;
    case 'l': // dump pressure - flow log
      dump_pf_log() ;
      break ;
    case 'm':
      reset_pf_log() ;
      break ;
    case 'q':
      for(i=0;i<100;i++) {
	val = in_p ;
	put_word(val) ; space() ;
	_delay_ms(100) ;
      }
      break ;
    case 'r': // read EE PROM word
      val = get_val() ; // address
      addr = val ;
      val2 = read_at_word(addr) ;
      put_word(val2) ; space() ; put_byte(val2 & 0xFF) ; crlf() ;
      break ;
    case 's': // read honeywell pressure sensor once
      val = honeywell_get_pressure() ;
      put_string_P(PSTR("IN_P = ")); 
      put_word(val) ;
      crlf() ;
      break ;
    case 't': // read nxp pressure sensor
      val = nxp_get_pressure() ;
      put_string_P(PSTR("NXP = ")); 
      put_word(val) ;
      crlf() ;
      break ;
    case 'u': // get input voltage
      val = get_adc(V12_ADC_CHAN) ;
      put_string_P(PSTR("V12 = ")); 
      put_word(val) ;
      crlf() ;
      break ;
    case 'v': // honeywell pressure sensor test
      put_word(honeywell_press_zero) ; crlf() ;
      val = get_adc(HONEYWELL_PRESS_ADC_CHAN) ;
      put_word(val) ; crlf() ;
      val = honeywell_get_pressure() ;
      put_word(val) ; crlf() ;
      for(i=0;i<100;i++) {
	cli() ;
	val = get_adc(HONEYWELL_PRESS_ADC_CHAN) ;
	sei() ;
	put_word(val) ; space() ;
	_delay_ms(100) ;
      }
      break ;
    case 'w': // test watchdog
      watchdog_test() ;
      break ;
    case 'x':
      val = get_val() ;
      put_string_P(PSTR("PWM = ")); put_word(val) ; crlf() ;
      set_pwm0(val) ;
      break ;
    case 'y': // read Siargo RAW
      val = get_adc(SIARGO_ADC_CHAN) ;
      put_word(val) ; crlf() ;
      put_string_P(PSTR("SZ =\n\r")) ; put_word(siargo_flow_zero) ;
      break ;
    case 'z': // read pressure sensor - repeatedly
      for(i=0;i<100;i++) {
	val = honeywell_get_pressure() ;
	put_word(val) ; space() ;
	_delay_ms(100) ;
      }
      break ;
    case 'A': // run test script
      run_script() ;
      break ;
    case 'B':
      put_string_P(PSTR("Building Table\r\n")); 
      build_f2p_tab(0) ;
      build_f2p_tab(1) ;
      break ;
    case 'C': // test control ** - fix later for 2 gas
      val = get_val() ; // flow
      val2 = get_val() ; // mflow
      //i = pid_ff(val, val2) ;
      i = 0 ;
      put_word(val) ; dot() ; put_word(val2) ; dot() ; put_word(i) ; crlf() ;
      break ;    
    case 'O': // one breath - logging on
      one_breath() ;
      break ;
    case 'D':
      val = get_val() ;
      dgain = val ;
      put_string_P(PSTR("DGain = "));
      put_word(dgain) ; print_divisor() ;
      break ;
    case 'E':
      dump_f2p_tab(0) ;
      dump_f2p_tab(1) ;
      break ;
    case 'F':
      flow_test(0) ;
      break ;
    case 'G':
      cl_flow_test(0) ;
      break ;
    case 'H':
      front_panel_lock = 1 ;
      break ;
    case 'I':
      val = get_val() ;
      igain = val ;
      put_string_P(PSTR("IGain = "));
      put_word(igain) ; print_divisor() ;
      break ;
    case 'J':
      val = get_val() ;
      set_leds(val) ;
      break ;
    case 'K':
      front_panel_lock = 0 ;
      break ;
    case 'L': // dump log
      val = get_val() ; // lines to dump
      if(val == 0) val = MAX_LOG ;
      dump_log(val) ;
      break ;
    case 'M':
      trise = get_val() ;
      if(trise < TICK) trise = TICK ;
      if(trise > 300) trise = 300 ;
      put_string_P(PSTR("T Rise = ")); put_word(trise) ; crlf() ;
      compute_derived() ;
      break ;
    case 'N':
      no_gain() ;
      break ;
    case 'P':
      val = get_val() ;
      pgain = val ;
      put_string_P(PSTR("PGain = "));
      put_word(pgain) ; print_divisor() ;
      break ;
    case 'Q':
      done = 1 ;
      break ;
    case 'R':
      val = get_flow(0) ;
      put_string_P(PSTR("Flow = "));
      put_word(val) ; crlf() ;
      break ;
    case 'S':
      pwm_sweep(100,0,1,0) ;
      break ;
    case 'T':
      pwm_sweep(100,1,1,0) ;
      break ;
    case 'U':
      log_step = get_val() ;
      put_string_P(PSTR("Log Step =")) ; put_byte(log_step) ; crlf() ;
      break ;
    case 'V':
      put_string_P(PSTR("Reading Table\n\r")) ;
      read_f2p_tab(0) ;
      read_f2p_tab(1) ;
      break ;
    case 'W':
      put_string_P(PSTR("Writing Table\n\r")) ;
      store_f2p_tab(0) ;
      store_f2p_tab(1) ;
      break ;
    case 'X':
      val = get_val() ;
      val2 = siargo_raw_to_mlps(val) ;
      put_word(val) ; dot() ; put_word(val2) ; crlf() ;
      break ;
    case 'Y':
      val = get_val() ;
      val2 = get_val() ;
      press_gain = val ;
      press_igain = val2 ;
      val = get_val() ;
      press_dgain = val ;
      put_string_P(PSTR("Press Gain = ")) ; put_word(press_gain) ; space() ;
      put_string_P(PSTR("Press I Gain = ")) ; put_word(press_igain) ; space() ;
      put_string_P(PSTR("Press D Gain = ")) ; put_word(press_dgain) ; crlf() ;
      break ;
    case 'Z':
      zero_all_pwm() ;
      _delay_ms(100) ;
      put_string_P(PSTR("Zero\n\r"));
      //siargo_set_zero() ;
      nxp_set_zero() ;
      honeywell_set_zero() ;
      break ;
    case '?':
      put_string_P(PSTR("DEBUG COMMANDS\n\r")) ;
      put_string_P(PSTR("a - test gtime wrap\n\r")) ;
      put_string_P(PSTR("b/c - sol3 on/off\n\r")) ;
      put_string_P(PSTR("d - save parameters\n\r")) ;
      put_string_P(PSTR("e <addr>  - read 16 words from EE Prom\n\r")) ;
      put_string_P(PSTR("f - flow to PWM\n\r")) ;
      put_string_P(PSTR("g - restore parameters\n\r")) ;
      put_string_P(PSTR("h - <addr> <data> write EE Prom\n\r")) ;
      put_string_P(PSTR("k - reset pressure flow log \n\r")) ;      
      put_string_P(PSTR("l - dump pressure flow log\n\r")) ;
      put_string_P(PSTR("m - reset pressure flow log\n\r")) ;
      put_string_P(PSTR("r <addr> - read EE Prom\n\r")) ;
      put_string_P(PSTR("w - test watchdog\n\r")) ;
      put_string_P(PSTR("x - set PWM\n\r")) ;
      put_string_P(PSTR("z - Read pressure sensor\n\r")) ;
      put_string_P(PSTR("A - run test script\n\r")) ;
      put_string_P(PSTR("B - build table\n\r")) ;
      put_string_P(PSTR("C f m - run controller\n\r")) ;      
      put_string_P(PSTR("D - set derivative gain\n\r")) ;
      put_string_P(PSTR("E - dump table\n\r")) ;
      put_string_P(PSTR("F - open-loop flow test\n\r")) ;
      put_string_P(PSTR("G - closed-loop flow test\n\r")) ;
      put_string_P(PSTR("H - lock out front panel\n\r")) ;
      put_string_P(PSTR("I - set integral gain\n\r")) ;
      put_string_P(PSTR("J - set LEDs\n\r")) ;
      put_string_P(PSTR("K - unlock front panel\n\r")) ;
      put_string_P(PSTR("L - dump log\n\r")) ;
      put_string_P(PSTR("M - set rise time\n\r")) ;
      put_string_P(PSTR("N - no gain\n\r")) ;
      put_string_P(PSTR("O - one breath\n\r")) ;            
      put_string_P(PSTR("P - set PGain\n\r")) ;      
      put_string_P(PSTR("Q - exit to normal UI\n\r")) ;      
      put_string_P(PSTR("R - read flow sensor\n\r")) ;      
      put_string_P(PSTR("S - sweep PWM and read flow\n\r")) ;
      put_string_P(PSTR("T - raw PWM sweep\n\r")) ;
      put_string_P(PSTR("U - set log step\n\r")) ;
      put_string_P(PSTR("V - read table from ROM\n\r")) ;
      put_string_P(PSTR("W - write table to ROM\n\r")) ;
      put_string_P(PSTR("X - test raw to mlps conversion\n\r")) ;
      put_string_P(PSTR("Y - set pressure gain\n\r")) ;
      put_string_P(PSTR("Z - set sensor zero point\n\r")) ;
      put_string_P(PSTR("\n\r")) ;      
      put_string_P(PSTR("\n\r")) ;      
      break ;
    }
  } while(done == 0) ;
}
